// ConsoleApplication1.cpp : This file contains the 'main' function. Program execution begins and ends there.
//

#include <iostream>
#include <Windows.h>
#include <ntsecapi.h>
#include <ntstatus.h>
#include <Sddl.h>

constexpr int N_PAGES_TO_ALLOC = 1;

void InitLsaString(PLSA_UNICODE_STRING LsaString, const wchar_t* String)
{
    DWORD StringLength;

    if (String == NULL) {
        LsaString->Buffer = NULL;
        LsaString->Length = 0;
        LsaString->MaximumLength = 0;
        return;
    }

    STATUS_SUCCESS;

    StringLength = wcslen(String);
    LsaString->Buffer = (LPWSTR)String;
    LsaString->Length = (USHORT)StringLength * sizeof(WCHAR);
    LsaString->MaximumLength = (USHORT)(StringLength + 1) * sizeof(WCHAR);
}

NTSTATUS OpenPolicy(LPWSTR ServerName, DWORD DesiredAccess, PLSA_HANDLE PolicyHandle)
{
    LSA_OBJECT_ATTRIBUTES ObjectAttributes;
    LSA_UNICODE_STRING ServerString;
    PLSA_UNICODE_STRING Server = NULL;

    // 
    // Always initialize the object attributes to all zeroes.
    // 
    ZeroMemory(&ObjectAttributes, sizeof(ObjectAttributes));

    if (ServerName != NULL) {
        // 
        // Make a LSA_UNICODE_STRING out of the LPWSTR passed in
        // 
        InitLsaString(&ServerString, ServerName);
        Server = &ServerString;
    }

    // 
    // Attempt to open the policy.
    // 
    return LsaOpenPolicy(
        Server,
        &ObjectAttributes,
        DesiredAccess,
        PolicyHandle
    );
}


NTSTATUS SetPrivilegeOnAccount(LSA_HANDLE PolicyHandle, PSID AccountSid, const wchar_t* PrivilegeName, BOOL bEnable)
{
    LSA_UNICODE_STRING PrivilegeString;

    // 
    // Create a LSA_UNICODE_STRING for the privilege name.
    // 
    InitLsaString(&PrivilegeString, PrivilegeName);

    // 
    // grant or revoke the privilege, accordingly
    // 
    if (bEnable) {
        return LsaAddAccountRights(
            PolicyHandle,       // open policy handle
            AccountSid,         // target SID
            &PrivilegeString,   // privileges
            1                   // privilege count
        );
    }
    else {
        return LsaRemoveAccountRights(
            PolicyHandle,       // open policy handle
            AccountSid,         // target SID
            FALSE,              // do not disable all rights
            &PrivilegeString,   // privileges
            1                   // privilege count
        );
    }
}

int main()
{
    HANDLE hToken = NULL;

    if (!OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken))
    {
        printf("OpenProcessToken failed. GetLastError returned: %d\n", GetLastError());
        return -1;
    }

    DWORD dwBufferSize = 0;

    // Probe the buffer size reqired for PTOKEN_USER structure
    if (!GetTokenInformation(hToken, TokenUser, NULL, 0, &dwBufferSize) &&
        (GetLastError() != ERROR_INSUFFICIENT_BUFFER))
    {
        printf("GetTokenInformation failed. GetLastError returned: %d\n", GetLastError());

        // Cleanup
        CloseHandle(hToken);
        hToken = NULL;

        return -1;
    }

    PTOKEN_USER pTokenUser = (PTOKEN_USER)malloc(dwBufferSize);

    // Retrieve the token information in a TOKEN_USER structure
    if (!GetTokenInformation(
        hToken,
        TokenUser,
        pTokenUser,
        dwBufferSize,
        &dwBufferSize))
    {
        printf("GetTokenInformation failed. GetLastError returned: %d\n", GetLastError());

        // Cleanup
        CloseHandle(hToken);
        hToken = NULL;

        return -1;
    }

    // Print SID string
    LPWSTR strsid;
    ConvertSidToStringSid(pTokenUser->User.Sid, &strsid);
    printf("User SID: %S\n", strsid);

    // Cleanup
    CloseHandle(hToken);
    hToken = NULL;

    NTSTATUS status;
    LSA_HANDLE policyHandle;

    if (status = OpenPolicy(NULL, POLICY_CREATE_ACCOUNT | POLICY_LOOKUP_NAMES, &policyHandle))
    {
        printf("OpenPolicy: %d\n", status);
        return -1;
    }

    // Add new privelege to the account
    if (status = SetPrivilegeOnAccount(policyHandle, pTokenUser->User.Sid, SE_LOCK_MEMORY_NAME, TRUE))
    {
        printf("OpenPSetPrivilegeOnAccountolicy: %d\n", status);
        return -1;
    }

    printf("SetPrivilegeOnAccount: OK\n");

    // Cleanup
    CloseHandle(hToken);
    hToken = NULL;
}
